"""
Copyright 2020 - Present Okta, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# AUTO-GENERATED! DO NOT EDIT FILE DIRECTLY
# SEE CONTRIBUTOR DOCUMENTATION

from urllib.parse import urlencode
from okta.models.user\
    import User
from okta.models.app_link\
    import AppLink
from okta.models.o_auth_2_client\
    import OAuth2Client
from okta.models.o_auth_2_scope_consent_grant\
    import OAuth2ScopeConsentGrant
from okta.models.o_auth_2_refresh_token\
    import OAuth2RefreshToken
from okta.models.user_credentials\
    import UserCredentials
from okta.models.forgot_password_response\
    import ForgotPasswordResponse
from okta.models.group\
    import Group
from okta.models.identity_provider\
    import IdentityProvider
from okta.models.user_activation_token\
    import UserActivationToken
from okta.models.temp_password\
    import TempPassword
from okta.models.reset_password_token\
    import ResetPasswordToken
from okta.models.response_links\
    import ResponseLinks
from okta.models.role\
    import Role
from okta.models.catalog_application\
    import CatalogApplication
from okta.models.subscription\
    import Subscription
from okta.utils import format_url
from okta.api_client import APIClient


class UserClient(APIClient):
    """
    A Client object for the User resource.
    """

    def __init__(self):
        self._base_url = ""

    async def list_users(
            self, query_params={},
            keep_empty_params=False
    ):
        """
        Lists users in your organization with pagination in mos
        t cases.  A subset of users can be returned that match
        a supported filter expression or search criteria.
        Args:
            query_params {dict}: Map of query parameters for request
            [query_params.q] {str}
            [query_params.after] {str}
            [query_params.limit] {str}
            [query_params.filter] {str}
            [query_params.search] {str}
            [query_params.sortBy] {str}
            [query_params.sortOrder] {str}
        Returns:
            list: Collection of User instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, User)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(User(item))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def create_user(
            self, create_user_request, query_params={},
            keep_empty_params=False
    ):
        """
        Creates a new user in your Okta organization with or wi
        thout credentials.
        Args:
            {create_user_request}
            query_params {dict}: Map of query parameters for request
            [query_params.activate] {str}
            [query_params.provider] {str}
            [query_params.nextLogin] {str}
        Returns:
            User
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        if isinstance(create_user_request, dict):
            body = create_user_request
        else:
            body = create_user_request.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, User)

        if error:
            return (None, response, error)

        try:
            result = User(
                response.get_body()
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def set_linked_object_for_user(
            self, associatedUserId, primaryRelationshipName, primaryUserId,
            keep_empty_params=False
    ):
        """
        Method for
        /api/v1/users/{associatedUserId}/linkedObjects/{primary
        RelationshipName}/{primaryUserId}
        Args:
            associated_user_id {str}
            primary_relationship_name {str}
            primary_user_id {str}
        """
        http_method = "put".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{associatedUserId}/linkedObjects
                /{primaryRelationshipName}/{primaryUserId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def deactivate_or_delete_user(
            self, userId, query_params={},
            keep_empty_params=False
    ):
        """
        Deletes a user permanently.  This operation can only be
        performed on users that have a `DEPROVISIONED` status.
        **This action cannot be recovered!**
        Args:
            user_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.sendEmail] {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_user(
            self, userId,
            keep_empty_params=False
    ):
        """
        Fetches a user from your Okta organization.
        Args:
            user_id {str}
        Returns:
            User
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, User)

        if error:
            return (None, response, error)

        try:
            result = User(
                response.get_body()
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def partial_update_user(
            self, userId, user, query_params={},
            keep_empty_params=False
    ):
        """
        Fetch a user by `id`, `login`, or `login shortname` if
        the short name is unambiguous.
        Args:
            user_id {str}
            {user}
            query_params {dict}: Map of query parameters for request
            [query_params.strict] {str}
        Returns:
            User
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        if isinstance(user, dict):
            body = user
        else:
            body = user.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, User)

        if error:
            return (None, response, error)

        try:
            result = User(
                response.get_body()
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def update_user(
            self, userId, user, query_params={},
            keep_empty_params=False
    ):
        """
        Update a user's profile and/or credentials using strict
        -update semantics.
        Args:
            user_id {str}
            {user}
            query_params {dict}: Map of query parameters for request
            [query_params.strict] {str}
        Returns:
            User
        """
        http_method = "put".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        if isinstance(user, dict):
            body = user
        else:
            body = user.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, User)

        if error:
            return (None, response, error)

        try:
            result = User(
                response.get_body()
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_app_links(
            self, userId,
            keep_empty_params=False
    ):
        """
        Fetches appLinks for all direct or indirect (via group
        membership) assigned applications.
        Args:
            user_id {str}
        Returns:
            list: Collection of AppLink instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/appLinks
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, AppLink)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(AppLink(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_user_clients(
            self, userId,
            keep_empty_params=False
    ):
        """
        Lists all client resources for which the specified user
        has grants or tokens.
        Args:
            user_id {str}
        Returns:
            list: Collection of OAuth2Client instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/clients
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OAuth2Client)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(OAuth2Client(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def revoke_grants_for_user_and_client(
            self, userId, clientId,
            keep_empty_params=False
    ):
        """
        Revokes all grants for the specified user and client
        Args:
            user_id {str}
            client_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/clients/{clientId}/grants
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def list_grants_for_user_and_client(
            self, userId, clientId, query_params={},
            keep_empty_params=False
    ):
        """
        Lists all grants for a specified user and client
        Args:
            user_id {str}
            client_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.expand] {str}
            [query_params.after] {str}
            [query_params.limit] {str}
        Returns:
            list: Collection of OAuth2ScopeConsentGrant instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/clients/{clientId}/grants
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OAuth2ScopeConsentGrant)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(OAuth2ScopeConsentGrant(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def revoke_tokens_for_user_and_client(
            self, userId, clientId,
            keep_empty_params=False
    ):
        """
        Revokes all refresh tokens issued for the specified Use
        r and Client.
        Args:
            user_id {str}
            client_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/clients/{clientId}/tokens
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def list_refresh_tokens_for_user_and_client(
            self, userId, clientId, query_params={},
            keep_empty_params=False
    ):
        """
        Lists all refresh tokens issued for the specified User
        and Client.
        Args:
            user_id {str}
            client_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.expand] {str}
            [query_params.after] {str}
            [query_params.limit] {str}
        Returns:
            list: Collection of OAuth2RefreshToken instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/clients/{clientId}/tokens
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OAuth2RefreshToken)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(OAuth2RefreshToken(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def revoke_token_for_user_and_client(
            self, userId, clientId, tokenId,
            keep_empty_params=False
    ):
        """
        Revokes the specified refresh token.
        Args:
            user_id {str}
            client_id {str}
            token_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/clients/{clientId}/tokens
                /{tokenId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_refresh_token_for_user_and_client(
            self, userId, clientId, tokenId, query_params={},
            keep_empty_params=False
    ):
        """
        Gets a refresh token issued for the specified User and
        Client.
        Args:
            user_id {str}
            client_id {str}
            token_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.expand] {str}
            [query_params.limit] {str}
            [query_params.after] {str}
        Returns:
            OAuth2RefreshToken
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/clients/{clientId}/tokens
                /{tokenId}
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OAuth2RefreshToken)

        if error:
            return (None, response, error)

        try:
            result = OAuth2RefreshToken(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def change_password(
            self, userId, change_password_request, query_params={},
            keep_empty_params=False
    ):
        """
        Changes a user's password by validating the user's curr
        ent password. This operation can only be performed on u
        sers in `STAGED`, `ACTIVE`, `PASSWORD_EXPIRED`, or `REC
        OVERY` status that have a valid password credential
        Args:
            user_id {str}
            {change_password_request}
            query_params {dict}: Map of query parameters for request
            [query_params.strict] {str}
        Returns:
            UserCredentials
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/credentials/change_password
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        if isinstance(change_password_request, dict):
            body = change_password_request
        else:
            body = change_password_request.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, UserCredentials)

        if error:
            return (None, response, error)

        try:
            result = UserCredentials(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def change_recovery_question(
            self, userId, user_credentials,
            keep_empty_params=False
    ):
        """
        Changes a user's recovery question & answer credential
        by validating the user's current password.  This operat
        ion can only be performed on users in **STAGED**, **ACT
        IVE** or **RECOVERY** `status` that have a valid passwo
        rd credential
        Args:
            user_id {str}
            {user_credentials}
        Returns:
            UserCredentials
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/credentials
                /change_recovery_question
            """)

        if isinstance(user_credentials, dict):
            body = user_credentials
        else:
            body = user_credentials.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, UserCredentials)

        if error:
            return (None, response, error)

        try:
            result = UserCredentials(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def forgot_password_generate_one_time_token(
            self, userId, query_params={},
            keep_empty_params=False
    ):
        """
        Generates a one-time token (OTT) that can be used to re
        set a user's password
        Args:
            user_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.sendEmail] {str}
        Returns:
            ForgotPasswordResponse
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/credentials/forgot_password
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, ForgotPasswordResponse)

        if error:
            return (None, response, error)

        try:
            result = ForgotPasswordResponse(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def forgot_password_set_new_password(
            self, userId, user_credentials, query_params={},
            keep_empty_params=False
    ):
        """
        Sets a new password for a user by validating the user's
        answer to their current recovery question
        Args:
            user_id {str}
            {user_credentials}
            query_params {dict}: Map of query parameters for request
            [query_params.sendEmail] {str}
        Returns:
            ForgotPasswordResponse
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/credentials/forgot_password
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        if isinstance(user_credentials, dict):
            body = user_credentials
        else:
            body = user_credentials.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, ForgotPasswordResponse)

        if error:
            return (None, response, error)

        try:
            result = ForgotPasswordResponse(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def revoke_user_grants(
            self, userId,
            keep_empty_params=False
    ):
        """
        Revokes all grants for a specified user
        Args:
            user_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/grants
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def list_user_grants(
            self, userId, query_params={},
            keep_empty_params=False
    ):
        """
        Lists all grants for the specified user
        Args:
            user_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.scopeId] {str}
            [query_params.expand] {str}
            [query_params.after] {str}
            [query_params.limit] {str}
        Returns:
            list: Collection of OAuth2ScopeConsentGrant instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/grants
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OAuth2ScopeConsentGrant)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(OAuth2ScopeConsentGrant(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def revoke_user_grant(
            self, userId, grantId,
            keep_empty_params=False
    ):
        """
        Revokes one grant for a specified user
        Args:
            user_id {str}
            grant_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/grants/{grantId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_user_grant(
            self, userId, grantId, query_params={},
            keep_empty_params=False
    ):
        """
        Gets a grant for the specified user
        Args:
            user_id {str}
            grant_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.expand] {str}
        Returns:
            OAuth2ScopeConsentGrant
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/grants/{grantId}
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OAuth2ScopeConsentGrant)

        if error:
            return (None, response, error)

        try:
            result = OAuth2ScopeConsentGrant(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_user_groups(
            self, userId,
            keep_empty_params=False
    ):
        """
        Fetches the groups of which the user is a member.
        Args:
            user_id {str}
        Returns:
            list: Collection of Group instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/groups
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Group)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(Group(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_user_identity_providers(
            self, userId,
            keep_empty_params=False
    ):
        """
        Lists the IdPs associated with the user.
        Args:
            user_id {str}
        Returns:
            list: Collection of IdentityProvider instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/idps
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, IdentityProvider)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(IdentityProvider(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def activate_user(
            self, userId, query_params={},
            keep_empty_params=False
    ):
        """
        Activates a user.  This operation can only be performed
        on users with a `STAGED` status.  Activation of a user
        is an asynchronous operation. The user will have the `
        transitioningToStatus` property with a value of `ACTIVE
        ` during activation to indicate that the user hasn't co
        mpleted the asynchronous operation.  The user will have
        a status of `ACTIVE` when the activation process is co
        mplete.
        Args:
            user_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.sendEmail] {str}
        Returns:
            UserActivationToken
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/lifecycle/activate
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, UserActivationToken)

        if error:
            return (None, response, error)

        try:
            result = UserActivationToken(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def deactivate_user(
            self, userId, query_params={},
            keep_empty_params=False
    ):
        """
        Deactivates a user. This operation can only be performe
        d on users that do not have a `DEPROVISIONED` status. W
        hile the asynchronous operation (triggered by HTTP head
        er `Prefer: respond-async`) is proceeding the user's `t
        ransitioningToStatus` property is `DEPROVISIONED`. The
        user's status is `DEPROVISIONED` when the deactivation
        process is complete.
        Args:
            user_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.sendEmail] {str}
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/lifecycle/deactivate
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def expire_password(
            self, userId,
            keep_empty_params=False
    ):
        """
        This operation transitions the user to the status of `P
        ASSWORD_EXPIRED` so that the user is required to change
        their password at their next login.
        Args:
            user_id {str}
        Returns:
            User
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/lifecycle
                /expire_password?tempPassword=false
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, User)

        if error:
            return (None, response, error)

        try:
            result = User(
                response.get_body()
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def expire_password_and_get_temporary_password(
            self, userId,
            keep_empty_params=False
    ):
        """
        This operation transitions the user to the status of `P
        ASSWORD_EXPIRED` and the user's password is reset to a
        temporary password that is returned.
        Args:
            user_id {str}
        Returns:
            TempPassword
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/lifecycle
                /expire_password?tempPassword=true
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, TempPassword)

        if error:
            return (None, response, error)

        try:
            result = TempPassword(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def reactivate_user(
            self, userId, query_params={},
            keep_empty_params=False
    ):
        """
        Reactivates a user.  This operation can only be perform
        ed on users with a `PROVISIONED` status.  This operatio
        n restarts the activation workflow if for some reason t
        he user activation was not completed when using the act
        ivationToken from [Activate User](#activate-user).
        Args:
            user_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.sendEmail] {str}
        Returns:
            UserActivationToken
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/lifecycle/reactivate
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, UserActivationToken)

        if error:
            return (None, response, error)

        try:
            result = UserActivationToken(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def reset_factors(
            self, userId,
            keep_empty_params=False
    ):
        """
        This operation resets all factors for the specified use
        r. All MFA factor enrollments returned to the unenrolle
        d state. The user's status remains ACTIVE. This link is
        present only if the user is currently enrolled in one
        or more MFA factors.
        Args:
            user_id {str}
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/lifecycle/reset_factors
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def reset_password(
            self, userId, query_params={},
            keep_empty_params=False
    ):
        """
        Generates a one-time token (OTT) that can be used to re
        set a user's password.  The OTT link can be automatical
        ly emailed to the user or returned to the API caller an
        d distributed using a custom flow.
        Args:
            user_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.sendEmail] {str}
        Returns:
            ResetPasswordToken
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/lifecycle/reset_password
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, ResetPasswordToken)

        if error:
            return (None, response, error)

        try:
            result = ResetPasswordToken(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def suspend_user(
            self, userId,
            keep_empty_params=False
    ):
        """
        Suspends a user.  This operation can only be performed
        on users with an `ACTIVE` status.  The user will have a
        status of `SUSPENDED` when the process is complete.
        Args:
            user_id {str}
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/lifecycle/suspend
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def unlock_user(
            self, userId,
            keep_empty_params=False
    ):
        """
        Unlocks a user with a `LOCKED_OUT` status and returns t
        hem to `ACTIVE` status.  Users will be able to login wi
        th their current password.
        Args:
            user_id {str}
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/lifecycle/unlock
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def unsuspend_user(
            self, userId,
            keep_empty_params=False
    ):
        """
        Unsuspends a user and returns them to the `ACTIVE` stat
        e.  This operation can only be performed on users that
        have a `SUSPENDED` status.
        Args:
            user_id {str}
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/lifecycle/unsuspend
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def remove_linked_object_for_user(
            self, userId, relationshipName,
            keep_empty_params=False
    ):
        """
        Delete linked objects for a user, relationshipName can
        be ONLY a primary relationship name
        Args:
            user_id {str}
            relationship_name {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/linkedObjects
                /{relationshipName}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_linked_objects_for_user(
            self, userId, relationshipName, query_params={},
            keep_empty_params=False
    ):
        """
        Get linked objects for a user, relationshipName can be
        a primary or associated relationship name
        Args:
            user_id {str}
            relationship_name {str}
            query_params {dict}: Map of query parameters for request
            [query_params.after] {str}
            [query_params.limit] {str}
        Returns:
            list: Collection of ResponseLinks instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/linkedObjects
                /{relationshipName}
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, ResponseLinks)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(ResponseLinks(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_assigned_roles_for_user(
            self, userId, query_params={},
            keep_empty_params=False
    ):
        """
        Lists all roles assigned to a user.
        Args:
            user_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.expand] {str}
        Returns:
            list: Collection of Role instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/roles
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Role)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(Role(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def assign_role_to_user(
            self, userId, assign_role_request, query_params={},
            keep_empty_params=False
    ):
        """
        Assigns a role to a user.
        Args:
            user_id {str}
            {assign_role_request}
            query_params {dict}: Map of query parameters for request
            [query_params.disableNotifications] {str}
        Returns:
            Role
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/roles
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        if isinstance(assign_role_request, dict):
            body = assign_role_request
        else:
            body = assign_role_request.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Role)

        if error:
            return (None, response, error)

        try:
            result = Role(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def remove_role_from_user(
            self, userId, roleId,
            keep_empty_params=False
    ):
        """
        Unassigns a role from a user.
        Args:
            user_id {str}
            role_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/roles/{roleId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_user_role(
            self, userId, roleId,
            keep_empty_params=False
    ):
        """
        Gets role that is assigne to user.
        Args:
            user_id {str}
            role_id {str}
        Returns:
            Role
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/roles/{roleId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Role)

        if error:
            return (None, response, error)

        try:
            result = Role(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def list_app_targets_for_application_admin_role_for_user(
            self, userId, roleId, query_params={},
            keep_empty_params=False
    ):
        """
        Lists all App targets for an `APP_ADMIN` Role assigned
        to a User. This methods return list may include full Ap
        plications or Instances. The response for an instance w
        ill have an `ID` value, while Application will not have
        an ID.
        Args:
            user_id {str}
            role_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.after] {str}
            [query_params.limit] {str}
        Returns:
            list: Collection of CatalogApplication instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/roles/{roleId}/targets
                /catalog/apps
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, CatalogApplication)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(CatalogApplication(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def add_all_apps_as_target_to_role(
            self, userId, roleId,
            keep_empty_params=False
    ):
        """
        Args:
            user_id {str}
            role_id {str}
        """
        http_method = "put".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/roles/{roleId}/targets
                /catalog/apps
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def remove_app_target_from_application_admin_role_for_user(
            self, userId, roleId, appName,
            keep_empty_params=False
    ):
        """
        Args:
            user_id {str}
            role_id {str}
            app_name {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/roles/{roleId}/targets
                /catalog/apps/{appName}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def add_application_target_to_admin_role_for_user(
            self, userId, roleId, appName,
            keep_empty_params=False
    ):
        """
        Args:
            user_id {str}
            role_id {str}
            app_name {str}
        """
        http_method = "put".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/roles/{roleId}/targets
                /catalog/apps/{appName}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def remove_application_target_from_administrator_role_for_user(
            self, userId, roleId, appName, applicationId,
            keep_empty_params=False
    ):
        """
        Remove App Instance Target to App Administrator Role gi
        ven to a User
        Args:
            user_id {str}
            role_id {str}
            app_name {str}
            application_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/roles/{roleId}/targets
                /catalog/apps/{appName}/{applicationId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def add_application_target_to_app_admin_role_for_user(
            self, userId, roleId, appName, applicationId,
            keep_empty_params=False
    ):
        """
        Add App Instance Target to App Administrator Role given
        to a User
        Args:
            user_id {str}
            role_id {str}
            app_name {str}
            application_id {str}
        """
        http_method = "put".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/roles/{roleId}/targets
                /catalog/apps/{appName}/{applicationId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def list_group_targets_for_role(
            self, userId, roleId, query_params={},
            keep_empty_params=False
    ):
        """
        Args:
            user_id {str}
            role_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.after] {str}
            [query_params.limit] {str}
        Returns:
            list: Collection of Group instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/roles/{roleId}/targets
                /groups
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Group)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(Group(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def remove_group_target_from_role(
            self, userId, roleId, groupId,
            keep_empty_params=False
    ):
        """
        Args:
            user_id {str}
            role_id {str}
            group_id {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/roles/{roleId}/targets
                /groups/{groupId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def add_group_target_to_role(
            self, userId, roleId, groupId,
            keep_empty_params=False
    ):
        """
        Args:
            user_id {str}
            role_id {str}
            group_id {str}
        """
        http_method = "put".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/roles/{roleId}/targets
                /groups/{groupId}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def clear_user_sessions(
            self, userId, query_params={},
            keep_empty_params=False
    ):
        """
        Removes all active identity provider sessions. This for
        ces the user to authenticate on the next operation. Opt
        ionally revokes OpenID Connect and OAuth refresh and ac
        cess tokens issued to the user.
        Args:
            user_id {str}
            query_params {dict}: Map of query parameters for request
            [query_params.oauthTokens] {str}
        """
        http_method = "delete".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/sessions
            """)
        if query_params:
            encoded_query_params = urlencode(query_params)
            api_url += f"/?{encoded_query_params}"

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def list_user_subscriptions(
            self, userId,
            keep_empty_params=False
    ):
        """
        List subscriptions of a User. Only lists subscriptions
        for current user. An AccessDeniedException message is s
        ent if requests are made from other users.
        Args:
            user_id {str}
        Returns:
            list: Collection of Subscription instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/subscriptions
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Subscription)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(Subscription(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def get_user_subscription_by_notification_type(
            self, userId, notificationType,
            keep_empty_params=False
    ):
        """
        Get the subscriptions of a User with a specific notific
        ation type. Only gets subscriptions for current user. A
        n AccessDeniedException message is sent if requests are
        made from other users.
        Args:
            user_id {str}
            notification_type {str}
        Returns:
            Subscription
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/users/{userId}/subscriptions
                /{notificationType}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, Subscription)

        if error:
            return (None, response, error)

        try:
            result = Subscription(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)



import pytest
from tests.mocks import MockOktaClient
import okta.models as models
from okta.constants import DATETIME_FORMAT
import datetime as dt


class TestLogEventsResource:
    """
    Integration Tests for the Log Events Resource
    """
    SDK_PREFIX = "python_sdk"

    @pytest.mark.asyncio
    @pytest.mark.skip
    async def test_get_logs(self):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp-GetLogs"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        try:
            # Create App in org
            app, _, err = await client.create_application(bookmark_app_obj)
            assert err is None
            assert isinstance(app, models.Application)
            assert isinstance(app, models.BookmarkApplication)

            logs, _, err = await client.get_logs()
            assert err is None
            assert logs is not None
            assert isinstance(logs, list)
            if logs[0]:
                assert isinstance(logs[0], models.LogEvent)

        finally:
            errors = []
            # Deactivate & Delete created app
            try:
                _, err = await client.deactivate_application(app.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.delete_application(app.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.asyncio
    @pytest.mark.skip
    async def test_get_logs_polling(self):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp-GetLogs"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        try:
            # Create App in org
            app, _, err = await client.create_application(bookmark_app_obj)
            assert err is None
            assert isinstance(app, models.Application)
            assert isinstance(app, models.BookmarkApplication)

            now = dt.datetime.now(dt.timezone.utc)
            now = now.replace(microsecond=0)
            one_minute_after = now + dt.timedelta(minutes=1)

            log_query_params = {
                "sortOrder": "ASCENDING",
                "until": one_minute_after.strftime(DATETIME_FORMAT)
            }

            logs, _, err = await client.get_logs(log_query_params)
            assert err is None
            assert logs is not None
            assert isinstance(logs, list)
            if logs[0]:
                assert isinstance(logs[0], models.LogEvent)

        finally:
            errors = []
            # Deactivate & Delete created app
            try:
                _, err = await client.deactivate_application(app.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.delete_application(app.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.asyncio
    @pytest.mark.skip
    async def test_get_logs_bounded(self):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Bookmark Application Object
        APP_URL = "https://example.com/bookmark.htm"
        APP_LABEL = "AddBookmarkApp-GetLogs"
        app_settings_app = models.BookmarkApplicationSettingsApplication({
            "requestIntegration": False,
            "url": APP_URL
        })
        app_settings = models.BookmarkApplicationSettings({
            "app": app_settings_app
        })
        bookmark_app_obj = models.BookmarkApplication({
            "label": APP_LABEL,
            "settings": app_settings
        })

        try:
            # Create App in org
            app, _, err = await client.create_application(bookmark_app_obj)
            assert err is None
            assert isinstance(app, models.Application)
            assert isinstance(app, models.BookmarkApplication)

            # Retrieve logs
            now = dt.datetime.now(dt.timezone.utc)
            now = now.replace(microsecond=0)
            one_hour_before = now - dt.timedelta(hours=1)

            log_query_params = {
                "since": one_hour_before.strftime(DATETIME_FORMAT),
                "until": now.strftime(DATETIME_FORMAT)
            }
            logs, _, err = await client.get_logs(log_query_params)
            assert err is None
            assert logs is not None
            assert isinstance(logs, list)
            if logs[0]:
                assert isinstance(logs[0], models.LogEvent)

        finally:
            errors = []
            # Deactivate & Delete created app
            try:
                _, err = await client.deactivate_application(app.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.delete_application(app.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0



from okta import helpers


def test_to_snake_case():
    assert helpers.to_snake_case('string') == 'string'
    assert helpers.to_snake_case('CamelCaseStr') == 'camel_case_str'
    assert helpers.to_snake_case('lowerCamelCaseStr') == 'lower_camel_case_str'
    assert helpers.to_snake_case('snake_case_str') == 'snake_case_str'


def test_to_lower_camel_case():
    assert helpers.to_lower_camel_case('') == ''
    assert helpers.to_lower_camel_case('string') == 'string'
    assert helpers.to_lower_camel_case('CamelCaseStr') == 'camelCaseStr'
    assert helpers.to_lower_camel_case('lowerCamelCaseStr') == 'lowerCamelCaseStr'
    assert helpers.to_lower_camel_case('snake_case_str') == 'snakeCaseStr'
    assert helpers.to_lower_camel_case('_start_with_underscore') == 'StartWithUnderscore'
    assert helpers.to_lower_camel_case('__start_with_double_underscore') == 'StartWithDoubleUnderscore'
    assert helpers.to_lower_camel_case('__double_underscores__') == 'DoubleUnderscores'



import logging


LOG_FORMAT = \
    '%(asctime)s - %(name)s - %(module)s - %(levelname)s - %(message)s'


def setup_logging(log_level=logging.INFO):
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(log_level)
    log_formatter = logging.Formatter(LOG_FORMAT)
    stream_handler.setFormatter(log_formatter)

    logger = logging.getLogger('okta-sdk-python')
    logger.addHandler(stream_handler)
    logger.setLevel(logging.DEBUG)

    # disable logger by default
    logger.disabled = True



import pytest
import okta.models as models
from tests.mocks import MockOktaClient
from okta.errors.okta_api_error import OktaAPIError
from http import HTTPStatus


class TestTemplatesResource:
    """
    Integration Tests for the Templates Resource.
    """
    SDK_PREFIX = "python_sdk"

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_sms_template(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create SMS Template
        sms_translations = models.SmsTemplateTranslations()
        sms_translations.es = \
            "${org.name}: el código de verificación es ${code}"
        sms_translations.fr = \
            "${org.name}: votre code de vérification est ${code}"
        sms_translations.it = "${org.name}: il codice di verifica è ${code}"

        sms_template_model = models.SmsTemplate({
            "name": f"{TestTemplatesResource.SDK_PREFIX}-test-template",
            "type": models.SmsTemplateType.SMS_VERIFY_CODE,
            "template": "${org.name}: your verification code is ${code}",
            "translations": sms_translations
        })

        try:
            created_template, _, err = await client.\
                create_sms_template(sms_template_model)
            assert err is None
            assert isinstance(created_template, models.SmsTemplate)
            assert created_template.name == sms_template_model.name
            assert created_template.type == sms_template_model.type
            assert created_template.template == sms_template_model.template

            # Retrieve
            retrieved_template, _, err = await client.\
                get_sms_template(created_template.id)
            assert err is None
            assert retrieved_template.id == created_template.id
            assert retrieved_template.name == created_template.name
            assert retrieved_template.template == created_template.template
            assert retrieved_template.translations.es == sms_translations.es
            assert retrieved_template.translations.it == sms_translations.it
            assert retrieved_template.translations.fr == sms_translations.fr

        finally:
            # Delete
            _, err = await client.delete_sms_template(created_template.id)
            assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_list_sms_templates(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create SMS Template
        sms_translations = models.SmsTemplateTranslations()
        sms_translations.es = \
            "${org.name}: el código de verificación es ${code}"
        sms_translations.fr = \
            "${org.name}: votre code de vérification est ${code}"
        sms_translations.it = "${org.name}: il codice di verifica è ${code}"

        sms_template_model = models.SmsTemplate({
            "name": f"{TestTemplatesResource.SDK_PREFIX}-test-template",
            "type": models.SmsTemplateType.SMS_VERIFY_CODE,
            "template": "${org.name}: your verification code is ${code}",
            "translations": sms_translations
        })

        try:
            created_template, _, err = await client.\
                create_sms_template(sms_template_model)
            assert err is None
            assert isinstance(created_template, models.SmsTemplate)
            assert created_template.name == sms_template_model.name
            assert created_template.type == sms_template_model.type
            assert created_template.template == sms_template_model.template
            assert created_template.translations.es == sms_translations.es
            assert created_template.translations.it == sms_translations.it
            assert created_template.translations.fr == sms_translations.fr

            # List
            sms_templates, _, err = await client.list_sms_templates()
            assert err is None
            assert isinstance(sms_templates, list)
            assert next((template for template in sms_templates
                         if template.id == created_template.id), None) is not None

        finally:
            # Delete
            _, err = await client.delete_sms_template(created_template.id)
            assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_update_sms_template(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create SMS Template
        sms_translations = models.SmsTemplateTranslations()
        sms_translations.es = \
            "${org.name}: el código de verificación es ${code}"
        sms_translations.fr = \
            "${org.name}: votre code de vérification est ${code}"
        sms_translations.it = "${org.name}: il codice di verifica è ${code}"

        sms_template_model = models.SmsTemplate({
            "name": f"{TestTemplatesResource.SDK_PREFIX}-test-template",
            "type": models.SmsTemplateType.SMS_VERIFY_CODE,
            "template": "${org.name}: your verification code is ${code}",
            "translations": sms_translations
        })

        try:
            created_template, _, err = await client.\
                create_sms_template(sms_template_model)
            assert err is None
            assert isinstance(created_template, models.SmsTemplate)
            assert created_template.name == sms_template_model.name
            assert created_template.type == sms_template_model.type
            assert created_template.template == sms_template_model.template

            # Retrieve
            retrieved_template, _, err = await client.\
                get_sms_template(created_template.id)
            assert err is None
            assert retrieved_template.id == created_template.id
            assert retrieved_template.name == created_template.name
            assert retrieved_template.template == created_template.template
            assert retrieved_template.translations.es == sms_translations.es
            assert retrieved_template.translations.it == sms_translations.it
            assert retrieved_template.translations.fr == sms_translations.fr

            # Update
            created_template.name = sms_template_model.name + "UPDATE"
            updated_template, _, err = await client.\
                update_sms_template(created_template.id, created_template)
            assert err is None
            assert updated_template.id == created_template.id
            assert updated_template.name == sms_template_model.name + "UPDATE"

            # Retrieve
            retrieved_template, _, err = await client.\
                get_sms_template(created_template.id)
            assert err is None
            assert retrieved_template.id == created_template.id
            assert retrieved_template.name == updated_template.name
            assert retrieved_template.template == created_template.template
            assert retrieved_template.translations.es == sms_translations.es
            assert retrieved_template.translations.it == sms_translations.it
            assert retrieved_template.translations.fr == sms_translations.fr

        finally:
            # Delete
            _, err = await client.delete_sms_template(created_template.id)
            assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_partial_update_sms_template(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create SMS Template
        sms_translations = models.SmsTemplateTranslations()
        sms_translations.es = \
            "${org.name}: el código de verificación es ${code}"
        sms_translations.fr = \
            "${org.name}: votre code de vérification est ${code}"
        sms_translations.it = "${org.name}: il codice di verifica è ${code}"

        sms_template_model = models.SmsTemplate({
            "name": f"{TestTemplatesResource.SDK_PREFIX}-test-template",
            "type": models.SmsTemplateType.SMS_VERIFY_CODE,
            "template": "${org.name}: your verification code is ${code}",
            "translations": sms_translations
        })

        try:
            created_template, _, err = await client.\
                create_sms_template(sms_template_model)
            assert err is None
            assert isinstance(created_template, models.SmsTemplate)
            assert created_template.name == sms_template_model.name
            assert created_template.type == sms_template_model.type
            assert created_template.template == sms_template_model.template

            # Retrieve
            retrieved_template, _, err = await client.\
                get_sms_template(created_template.id)
            assert err is None
            assert retrieved_template.id == created_template.id
            assert retrieved_template.name == created_template.name
            assert retrieved_template.template == created_template.template
            assert retrieved_template.translations.es == sms_translations.es
            assert retrieved_template.translations.it == sms_translations.it
            assert retrieved_template.translations.fr == sms_translations.fr

            # Update
            temporary_template = models.SmsTemplate({
                "translations": models.SmsTemplateTranslations({
                    "de":  "${org.name}: ihre bestätigungscode ist ${code}."
                })
            })
            updated_template, _, err = await client.\
                partial_update_sms_template(
                    created_template.id, temporary_template)
            assert err is None
            assert updated_template.id == created_template.id
            assert updated_template.name == created_template.name
            assert updated_template.type == created_template.type
            assert updated_template.translations.de ==\
                temporary_template.translations.de

            # Retrieve
            retrieved_template, _, err = await client.\
                get_sms_template(created_template.id)
            assert err is None
            assert retrieved_template.id == created_template.id
            assert retrieved_template.name == updated_template.name
            assert retrieved_template.template == created_template.template
            assert retrieved_template.translations.es == sms_translations.es
            assert retrieved_template.translations.it == sms_translations.it
            assert retrieved_template.translations.fr == sms_translations.fr
            assert retrieved_template.translations.de ==\
                temporary_template.translations.de

        finally:
            # Delete
            _, err = await client.delete_sms_template(created_template.id)
            assert err is None

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_delete_sms_template(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create SMS Template
        sms_translations = models.SmsTemplateTranslations()
        sms_translations.es = \
            "${org.name}: el código de verificación es ${code}"
        sms_translations.fr = \
            "${org.name}: votre code de vérification est ${code}"
        sms_translations.it = "${org.name}: il codice di verifica è ${code}"

        sms_template_model = models.SmsTemplate({
            "name": f"{TestTemplatesResource.SDK_PREFIX}-test-template",
            "type": models.SmsTemplateType.SMS_VERIFY_CODE,
            "template": "${org.name}: your verification code is ${code}",
            "translations": sms_translations
        })

        try:
            created_template, _, err = await client.\
                create_sms_template(sms_template_model)
            assert err is None
            assert isinstance(created_template, models.SmsTemplate)
            assert created_template.name == sms_template_model.name
            assert created_template.type == sms_template_model.type
            assert created_template.template == sms_template_model.template

            # Retrieve
            retrieved_template, _, err = await client.\
                get_sms_template(created_template.id)
            assert err is None
            assert retrieved_template.id == created_template.id
            assert retrieved_template.name == created_template.name
            assert retrieved_template.template == created_template.template
            assert retrieved_template.translations.es == sms_translations.es
            assert retrieved_template.translations.it == sms_translations.it
            assert retrieved_template.translations.fr == sms_translations.fr

            # Delete
            _, err = await client.delete_sms_template(created_template.id)
            assert err is None

            # Retrieve
            retrieved_template, resp, err = await client.\
                get_sms_template(created_template.id)
            assert err is not None
            assert isinstance(err, OktaAPIError)
            assert resp.get_status() == HTTPStatus.NOT_FOUND
            assert retrieved_template is None
        finally:
            try:
                _, err = await client.delete_sms_template(created_template.id)
            except Exception:
                pass



# flake8: noqa
"""
Copyright 2020 - Present Okta, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# AUTO-GENERATED! DO NOT EDIT FILE DIRECTLY
# SEE CONTRIBUTOR DOCUMENTATION

from okta.models.policy_rule_conditions\
    import PolicyRuleConditions
from okta.models import policy_rule_auth_context_condition\
    as policy_rule_auth_context_condition
from okta.models import policy_network_condition\
    as policy_network_condition
from okta.models import policy_people_condition\
    as policy_people_condition


class OktaSignOnPolicyRuleConditions(
    PolicyRuleConditions
):
    """
    A class for OktaSignOnPolicyRuleConditions objects.
    """

    def __init__(self, config=None):
        super().__init__(config)
        if config:
            if "authContext" in config:
                if isinstance(config["authContext"],
                              policy_rule_auth_context_condition.PolicyRuleAuthContextCondition):
                    self.auth_context = config["authContext"]
                elif config["authContext"] is not None:
                    self.auth_context = policy_rule_auth_context_condition.PolicyRuleAuthContextCondition(
                        config["authContext"]
                    )
                else:
                    self.auth_context = None
            else:
                self.auth_context = None
            if "network" in config:
                if isinstance(config["network"],
                              policy_network_condition.PolicyNetworkCondition):
                    self.network = config["network"]
                elif config["network"] is not None:
                    self.network = policy_network_condition.PolicyNetworkCondition(
                        config["network"]
                    )
                else:
                    self.network = None
            else:
                self.network = None
            if "people" in config:
                if isinstance(config["people"],
                              policy_people_condition.PolicyPeopleCondition):
                    self.people = config["people"]
                elif config["people"] is not None:
                    self.people = policy_people_condition.PolicyPeopleCondition(
                        config["people"]
                    )
                else:
                    self.people = None
            else:
                self.people = None
        else:
            self.auth_context = None
            self.network = None
            self.people = None

    def request_format(self):
        parent_req_format = super().request_format()
        current_obj_format = {
            "authContext": self.auth_context,
            "network": self.network,
            "people": self.people
        }
        parent_req_format.update(current_obj_format)
        return parent_req_format



from okta.http_client import HTTPClient
from okta.user_agent import UserAgent
from okta.oauth import OAuth
from okta.api_response import OktaAPIResponse
from okta.error_messages import ERROR_MESSAGE_429_MISSING_DATE_X_RESET
from okta.utils import convert_date_time_to_seconds
import time
from http import HTTPStatus
import json
import logging


logger = logging.getLogger('okta-sdk-python')


class RequestExecutor:
    """
    This class handles all of the requests sent by the Okta Client.
    """

    RETRY_COUNT_HEADER = 'X-Okta-Retry-Count'
    RETRY_FOR_HEADER = 'X-Okta-Retry-For'

    def __init__(self, config, cache, http_client=None):
        """
        Constructor for Request Executor object for Okta Client

        Arguments:
            config {dict} -- This dictionary contains the configuration
                             of the Request Executor
        """
        # Raise Value Error if numerical inputs are invalid (< 0)
        self._request_timeout = config["client"].get('requestTimeout', 0)
        if self._request_timeout < 0:
            raise ValueError(
                ("okta.client.requestTimeout provided as "
                 f"{self._request_timeout} but must be 0 (disabled) or "
                 "greater than zero"))
        self._max_retries = config["client"]["rateLimit"].get('maxRetries', 2)
        if self._max_retries < 0:
            raise ValueError(
                ("okta.client.rateLimit.maxRetries provided as "
                 f"{self._max_retries} but must be 0 (disabled) or "
                 "greater than zero"))
        # Setup other fields
        self._authorization_mode = config["client"]["authorizationMode"]
        self._base_url = config["client"]["orgUrl"]
        self._config = config
        self._cache = cache
        self._default_headers = {
            'User-Agent': UserAgent(config["client"].get("userAgent", None))
            .get_user_agent_string(),
            'Accept': "application/json"
        }

        # SSWS or Bearer header
        token_type = config["client"]["authorizationMode"]
        if token_type in ("SSWS", "Bearer"):
            self._default_headers['Authorization'] = (
                f"{token_type} {self._config['client']['token']}"
            )
        else:
            # OAuth
            self._oauth = OAuth(self, self._config)

        http_client_impl = http_client or HTTPClient
        self._http_client = http_client_impl({
            'requestTimeout': self._request_timeout,
            'headers': self._default_headers,
            'proxy': self._config["client"].get("proxy"),
            'sslContext': self._config["client"].get("sslContext"),
        })
        HTTPClient.raise_exception = \
            self._config['client'].get("raiseException", False)
        self._custom_headers = {}

    def clear_empty_params(self, body: dict):
        """
        Removes all key-value pairs where value is empty, i.e. empty string, list or dict.

        Args:
            body (dict): request body to be cleared

        Returns:
            dict: body without empty values.
        """
        if isinstance(body, dict):
            return {
                k: v
                for k, v in ((k, self.clear_empty_params(v)) for k, v in body.items())
                if v or v == 0 or v is False
            }
        if isinstance(body, list):
            return [v for v in map(self.clear_empty_params, body) if v or v == 0 or v is False]
        return body

    async def create_request(self, method: str, url: str, body: dict = None,
                             headers: dict = {}, form: dict = {}, oauth=False, keep_empty_params=False):
        """
        Creates request for request executor's HTTP client.

        Args:
            method (str): HTTP Method to be used
            url (str): URL to send request to
            body (dict, optional): Request body. Defaults to None.
            headers (dict, optional): Request headers. Defaults to {}.
            oauth: Should use oauth? Defaults to False.
            keep_empty_params: Should request body keep parameters with empty values? Defaults to False.

        Returns:
            dict, Exception: Tuple of Dictionary repr of HTTP request and
            exception raised during execution
        """
        # Base HTTP Request
        request = {
            "method": method
        }

        # Build request
        # Get predetermined headers and build URL
        headers.update(self._custom_headers)
        headers.update(self._default_headers)
        if self._config["client"]["orgUrl"] not in url:
            url = self._config["client"]["orgUrl"] + url

        # OAuth
        if self._authorization_mode == "PrivateKey" and not oauth:
            # check if access token exists
            if self._cache.contains("OKTA_ACCESS_TOKEN"):
                access_token = self._cache.get("OKTA_ACCESS_TOKEN")
            else:
                # if not, make one
                # Generate using private key provided
                access_token, error = await self._oauth.get_access_token()
                # return error if problem retrieving token
                if error:
                    return (None, error)

            # finally, add to header and cache
            headers.update({"Authorization": f"Bearer {access_token}"})
            self._cache.add("OKTA_ACCESS_TOKEN", access_token)

        # Add content type header if request body exists
        if body:
            headers.update({"Content-Type": "application/json"})
            if not keep_empty_params:
                body = self.clear_empty_params(body)

        # finish building request and return
        request["headers"] = headers
        request["url"] = url
        request["data"] = body
        request["form"] = form

        return (request, None)

    async def execute(self, request, response_type=None):
        """
        This function is the high level request execution method. Performs the
        API call and returns a formatted response object

        Args:
            request (dict): dictionary object containing request details

        Returns:
            (OktaAPIResponse, Exception): Response obj for the Okta API, Error
        """
        _, response, response_body, error = await self.fire_request(request)

        if error is not None:
            return (None, error)

        _, error = self._http_client.check_response_for_error(
            request["url"], response, response_body)

        return (
            OktaAPIResponse(self, request, response,
                            response_body, response_type),
            error
        )

    async def fire_request(self, request):
        """
        Send Request using HTTP Client

        Args:
            request (dict): HTTP request in dictionary format

        Returns:
            aiohttp.RequestInfo, aiohttp.ClientResponse, json, Exception: Tuple
            of request, response object, response json, and error if raised
        """
        # Retrieve URL from request and generate cache key
        url = request["url"]
        url_cache_key = self._cache.create_key(url)

        # Remove cache entry if not GET call
        method = request["method"]
        if method.upper() != "GET":
            self._cache.delete(url_cache_key)

        # check if in cache
        if not self._cache.contains(url_cache_key):
            # shoot request and return
            _, res_details, resp_body, error = await\
                self.fire_request_helper(request, 0, time.time())
            if error is not None:
                return (None, res_details, resp_body, error)

            # add to cache if not a list and valid response
            if method.upper() == "GET" and 200 <= res_details.status <= 299:
                try:
                    json_object = json.loads(resp_body)
                    if not isinstance(json_object, list):
                        self._cache.add(
                            url_cache_key, (res_details, resp_body))
                except Exception:
                    pass

            return (request, res_details, resp_body, error)

        # otherwise return cache response
        res_details, resp_body = self._cache.get(url_cache_key)
        return (request, res_details, resp_body, None)

    async def fire_request_helper(self, request, attempts, request_start_time):
        """
        Helper method to perform HTTP call with retries if needed

        Args:
            request (dict): HTTP request representation
            attempts (int): number of attempted HTTP calls so far
            request_start_time (float): original start time of request

        Returns:
            aiohttp.RequestInfo, aiohttp.ClientResponse, json, Exception: Tuple
            of request, response object, response json, and error if raised
        """
        # Get start request time
        current_req_start_time = time.time()
        max_retries = self._max_retries
        req_timeout = self._request_timeout

        if req_timeout > 0 and \
                (current_req_start_time - request_start_time) > req_timeout:
            # Timeout is hit for request
            return (None, None, None, Exception("Request Timeout exceeded."))

        # Execute request
        _, res_details, resp_body, error = \
            await self._http_client.send_request(request)
        # return immediately if request failed to launch (e.g. network is down, thus res_details is None)
        if res_details is None:
            return (None, None, None, error)

        headers = res_details.headers

        if attempts < max_retries and self.is_retryable_status(res_details.status):
            date_time = headers.get("Date", "")
            if date_time:
                date_time = convert_date_time_to_seconds(date_time)
            retry_limit_reset_headers = list(map(float, headers.getall(
                "X-Rate-Limit-Reset", [])))
            # header might be in lowercase, so check this too
            retry_limit_reset_headers.extend(list(map(float, headers.getall(
                "x-rate-limit-reset", []))))
            retry_limit_reset = min(retry_limit_reset_headers) if len(
                retry_limit_reset_headers) > 0 else None
            if not date_time or not retry_limit_reset:
                return (None, res_details, resp_body,
                        Exception(
                            ERROR_MESSAGE_429_MISSING_DATE_X_RESET
                        ))

            check_429 = self.is_too_many_requests(res_details.status, resp_body)
            if check_429:
                # backoff
                backoff_seconds = self.calculate_backoff(
                    retry_limit_reset, date_time)
                logger.info(f'Hit rate limit. Retry request in {backoff_seconds} seconds.')
                logger.debug(f'Value of retry_limit_reset: {retry_limit_reset}')
                logger.debug(f'Value of date_time: {date_time}')
                self.pause_for_backoff(backoff_seconds)
                if (current_req_start_time + backoff_seconds)\
                        - request_start_time > req_timeout and req_timeout > 0:
                    return (None, res_details, resp_body, resp_body)

            # Setup retry request
            attempts += 1
            request['headers'].update(
                {
                    RequestExecutor.RETRY_FOR_HEADER: headers.get(
                        "X-Okta-Request-Id", ""),
                    RequestExecutor.RETRY_COUNT_HEADER: str(attempts)
                }
            )

            _, res_details, resp_body, error = await self.fire_request_helper(
                request, attempts, request_start_time
            )
            if error:
                return (None, res_details, resp_body, error)

        return (request, res_details, resp_body, error)

    def is_retryable_status(self, status):
        """
        Checks if HTTP status is retryable.

        Retryable statuses: 429, 503, 504
        """
        return status is not None and status in (HTTPStatus.TOO_MANY_REQUESTS,
                                                 HTTPStatus.SERVICE_UNAVAILABLE,
                                                 HTTPStatus.GATEWAY_TIMEOUT)

    def is_too_many_requests(self, status, response):
        """
        Determines if HTTP request has been made too many times

        Args:
            status (int): HTTP response status code
            response (json): Response Body

        Returns:
            bool: Returns True if this request has been called too many times
        """
        return response is not None and status is not None\
            and status == HTTPStatus.TOO_MANY_REQUESTS

    def parse_response(self, request, response):
        pass

    def calculate_backoff(self, retry_limit_reset, date_time):
        return retry_limit_reset - date_time + 1

    def pause_for_backoff(self, backoff_time):
        time.sleep(backoff_time)

    def set_custom_headers(self, headers):
        self._custom_headers.update(headers)

    def set_session(self, session):
        self._http_client.set_session(session)

    def clear_custom_headers(self):
        self._custom_headers = {}

    def get_custom_headers(self):
        return self._custom_headers

    def get_default_headers(self):
        return self._default_headers



import os
import logging
import yaml
from okta.constants import _GLOBAL_YAML_PATH, _LOCAL_YAML_PATH
from flatdict import FlatDict


class ConfigSetter():
    """
    This class sets up the configuration for the Okta Client
    """

    # Configuration constants
    _OKTA = "OKTA"
    _DEFAULT_CONFIG = {
        "client": {
            "authorizationMode": '',
            "orgUrl": '',
            "token": '',
            "clientId": '',
            "scopes": '',
            "privateKey": '',
            "userAgent": '',
            "cache": {
                "enabled": '',
                "defaultTti": '',
                "defaultTtl": ''
            },
            "logging": {
                "enabled": '',
            },
            "proxy": {
                "port": "",
                "host": "",
                "username": "",
                "password": ""
            },
            "rateLimit": {
                "maxRetries": ''
            }
        },
        "testing": {
            "testingDisableHttpsCheck": ''
        }
    }

    def __init__(self):
        """
        Consructor for Configuration Setter class. Sets default config
        and checks for configuration settings to update config.
        """
        # Create configuration using default config
        self._config = ConfigSetter._DEFAULT_CONFIG
        # Update configuration
        self._update_config()

    def get_config(self):
        """
        Return Okta client configuration

        Returns:
            dict -- Dictionary containing the client configuration
        """
        return self._config

    def _prune_config(self, config):
        """
        This method cleans up the configuration object by removing fields
        with no value
        """
        # Flatten dictionary to account for nested dictionary
        flat_current_config = FlatDict(config, delimiter='_')
        # Iterate through keys and remove if value is still empty string
        for key in flat_current_config.keys():
            if flat_current_config.get(key) == '':
                del flat_current_config[key]

        return flat_current_config.as_dict()

    def _update_config(self):
        """
        Updates the configuration of the Okta Client by:
        1. Applying default values
        2. Checking for a global OKTA config YAML
        3. Checking for a local OKTA config YAML
        4. Checking for corresponding ENV variables
        """
        # apply default config values to config
        self._apply_default_values()
        # check if GLOBAL yaml exists, apply if true
        if (os.path.exists(_GLOBAL_YAML_PATH)):
            self._apply_yaml_config(_GLOBAL_YAML_PATH)
        # check if LOCAL yaml exists, apply if true
        if (os.path.exists(_LOCAL_YAML_PATH)):
            self._apply_yaml_config(_LOCAL_YAML_PATH)
        # apply existing environment variables
        self._apply_env_config()

    def _apply_default_values(self):
        """Apply default values to default client configuration
        """
        # Set defaults
        self._config["client"]["authorizationMode"] = "SSWS"
        self._config["client"]["connectionTimeout"] = 30
        self._config["client"]["cache"] = {
            "enabled": False,
            "defaultTtl": 300,
            "defaultTti": 300
        }
        self._config["client"]["logging"] = {
            "enabled": False,
            "logLevel": logging.INFO
        }
        self._config["client"]["userAgent"] = ""
        self._config["client"]["requestTimeout"] = 0
        self._config["client"]["rateLimit"] = {
            "maxRetries": 2
        }

        self._config["testing"]["testingDisableHttpsCheck"] = False

    def _apply_config(self, new_config: dict):
        """This method applies a config dictionary to the current config,
           overwriting values and adding new entries (if present).

        Arguments:
            config {dict} -- A dictionary of client configuration details
        """
        # Update current configuration with new configuration
        # Flatten both dictionaries to account for nested dictionary values
        flat_current_client = FlatDict(self._config['client'], delimiter='_')
        flat_current_testing = FlatDict(self._config['testing'], delimiter='_')

        flat_new_client = FlatDict(new_config.get('client', {}), delimiter='_')
        flat_new_testing = FlatDict(
            new_config.get('testing', {}), delimiter='_')
        flat_current_client.update(flat_new_client)
        flat_current_testing.update(flat_new_testing)
        # Update values in current config and unflatten
        self._config = {'client': flat_current_client.as_dict(),
                        'testing': flat_current_testing.as_dict()}

    def _apply_yaml_config(self, path: str):
        """This method applies a YAML configuration to the Okta Client Config

        Arguments:
            path {string} -- The filepath of the corresponding YAML file
        """
        # Start with empty config
        config = {}
        with open(path, 'r') as file:
            # Open file stream and attempt to load YAML
            config = yaml.load(file, Loader=yaml.SafeLoader)
        # Apply acquired config to configuration
        self._apply_config(config.get("okta", {}))

    def _apply_env_config(self):
        """
        This method checks the environment variables for any OKTA
        configuration parameters and applies them if available.
        """
        # Flatten current config and join with underscores
        # (for environment variable format)
        flattened_config = FlatDict(self._config, delimiter='_')
        flattened_keys = flattened_config.keys()

        # Create empty result config and populate
        updated_config = FlatDict({}, delimiter='_')

        # Go through keys and search for it in the environment vars
        # using the format described in the README
        for key in flattened_keys:
            env_key = ConfigSetter._OKTA + "_" + key.upper()
            env_value = os.environ.get(env_key, None)

            if env_value is not None:
                # If value is found, add to config
                if "scopes" in env_key.lower():
                    updated_config[key] = env_value.split(',')
                else:
                    updated_config[key] = env_value
            # apply to current configuration
        self._apply_config(updated_config.as_dict())



import aiohttp
import asyncio
import datetime
import time

from http import HTTPStatus
from multidict import MultiDict
from okta.client import Client as OktaClient
from okta.request_executor import RequestExecutor


def test_retry_count_header(monkeypatch):
    org_url = "https://test.okta.com"
    token = "TOKEN"
    config = {'orgUrl': org_url, 'token': token, 'rateLimit': {'maxRetries': 2}}
    client = OktaClient(config)

    class MockHTTPRequest():
        """
        async with aiohttp.request(**params) as response:
            return (response.request_info,
                    response,
                    await response.text(),
                    None)

        _, res_details, resp_body, error = \
        headers = res_details.headers
        """

        def __call__(self, **params):
            self.request_info = params
            self.headers = MultiDict({'Date': datetime.datetime.now(tz=datetime.timezone.utc).strftime('%a, %d %b %Y %H:%M:%S %Z'),
                                      'Content-Type': 'application/json',
                                      'X-Rate-Limit-Limit': 600,
                                      'X-Rate-Limit-Remaining': 599,
                                      'X-Rate-Limit-Reset': str(time.time())})
            self.url = params['url']
            self.content_type = 'application/json'
            self.links = ''
            self.text = MockHTTPRequest.mock_response_text
            self.status = HTTPStatus.TOO_MANY_REQUESTS
            return self

        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            pass

        @staticmethod
        async def mock_response_text():
            return '[{"embedded": null,' \
                   '"links": {"self": {"href": "https://test.okta.com/v1/users/test_id"}},' \
                   '"activated": "2021-01-01T00:00:00.000Z",' \
                   '"created": "2021-01-01T00:00:00.000Z",' \
                   '"credentials": null,' \
                   '"id": "test_id",' \
                   '"last_login": null,' \
                   '"profile": {"name": "test_name"},' \
                   '"status": null,' \
                   '"status_changed": null,' \
                   '"transitioning_to_status": null,' \
                   '"type": null}]'

    mock_http_request = MockHTTPRequest()
    monkeypatch.setattr(aiohttp.ClientSession, 'request', mock_http_request)
    res, resp_body, error = asyncio.run(client.list_users())
    # Check request was retried max times and header 'X-Okta-Retry-Count' was set properly
    assert mock_http_request.request_info['headers'].get('X-Okta-Retry-Count') == '2'


def test_is_retryable_status():
    org_url = "https://test.okta.com"
    token = "TOKEN"
    config = {'client': {'orgUrl': org_url,
                         'token': token,
                         'rateLimit': {},
                         'authorizationMode': None}}
    req_exec = RequestExecutor(config=config, cache=None)
    assert req_exec.is_retryable_status(HTTPStatus.TOO_MANY_REQUESTS)
    assert req_exec.is_retryable_status(HTTPStatus.SERVICE_UNAVAILABLE)
    assert req_exec.is_retryable_status(HTTPStatus.GATEWAY_TIMEOUT)
    assert not req_exec.is_retryable_status(HTTPStatus.FORBIDDEN)


def test_clear_empty_params():
    org_url = "https://test.okta.com"
    token = "TOKEN"
    config = {'client': {'orgUrl': org_url,
                         'token': token,
                         'rateLimit': {},
                         'authorizationMode': None}}
    req_exec = RequestExecutor(config=config, cache=None)

    body = {'int_value': 0,
            'str_value': '0',
            'empty_str_value': '',
            'list_value': [1,2,3],
            'empty_list_value': [],
            'dict_value': {'int_value': 0},
            'empty_dict_value': {},
            'nested_empty_dict_value': {'list_value': [1,2,3], 'empty_list_value': []},
            'nested_empty_list_value': {'empty_list_value': []}}

    cleared_body = {'int_value': 0,
                    'str_value': '0',
                    'list_value': [1,2,3],
                    'dict_value': {'int_value': 0},
                    'nested_empty_dict_value': {'list_value': [1,2,3]}}

    assert req_exec.clear_empty_params(body) == cleared_body



from okta.utils import convert_absolute_url_into_relative_url


def test_convert_absolute_url_into_relative_url():
    absolute_url = 'https://test.okta.com/api/v1/users'
    relative_url = '/api/v1/users'
    assert relative_url == convert_absolute_url_into_relative_url(absolute_url)

    absolute_url = 'https://test.okta.com/api/v1/users?activate=false'
    relative_url = '/api/v1/users?activate=false'
    assert relative_url == convert_absolute_url_into_relative_url(absolute_url)



import okta.models as models
from okta.constants import find_app_model

def test_find_app_model():
    expected = models.SamlApplication
    actual = find_app_model('SAML_1_1', 'office365')
    assert expected == actual

    expected = models.SwaApplication
    actual = find_app_model(None, 'template_swa')
    assert expected == actual

    expected = models.Application
    actual = find_app_model(None, 'unknown_application')
    assert expected == actual



"""
Copyright 2021 - Present Okta, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# AUTO-GENERATED! DO NOT EDIT FILE DIRECTLY
# SEE CONTRIBUTOR DOCUMENTATION

from okta.models.org_setting\
    import OrgSetting
from okta.models.org_contact_type_obj\
    import OrgContactTypeObj
from okta.models.org_contact_user\
    import OrgContactUser
from okta.models.org_preferences\
    import OrgPreferences
from okta.models.org_okta_communication_setting\
    import OrgOktaCommunicationSetting
from okta.models.org_okta_support_settings_obj\
    import OrgOktaSupportSettingsObj
from okta.utils import format_url
from okta.api_client import APIClient


class OrgClient(APIClient):
    """
    A Client object for the Org resource.
    """

    def __init__(self):
        self._base_url = ""

    async def get_org_settings(
            self,
            keep_empty_params=False
    ):
        """
        Get settings of your organization.
        Args:
        Returns:
            OrgSetting
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/org
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OrgSetting)

        if error:
            return (None, response, error)

        try:
            result = OrgSetting(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def partial_update_org_setting(
            self, org_setting,
            keep_empty_params=False
    ):
        """
        Partial update settings of your organization.
        Args:
            {org_setting}
        Returns:
            OrgSetting
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/org
            """)

        if isinstance(org_setting, dict):
            body = org_setting
        else:
            body = org_setting.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OrgSetting)

        if error:
            return (None, response, error)

        try:
            result = OrgSetting(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def update_org_setting(
            self, org_setting,
            keep_empty_params=False
    ):
        """
        Update settings of your organization.
        Args:
            {org_setting}
        Returns:
            OrgSetting
        """
        http_method = "put".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/org
            """)

        if isinstance(org_setting, dict):
            body = org_setting
        else:
            body = org_setting.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OrgSetting)

        if error:
            return (None, response, error)

        try:
            result = OrgSetting(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def get_org_contact_types(
            self,
            keep_empty_params=False
    ):
        """
        Gets Contact Types of your organization.
        Args:
        Returns:
            list: Collection of OrgContactTypeObj instances.
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/org/contacts
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OrgContactTypeObj)

        if error:
            return (None, response, error)

        try:
            result = []
            for item in response.get_body():
                result.append(OrgContactTypeObj(
                    self.form_response_body(item)
                    ))
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def get_org_contact_user(
            self, contactType,
            keep_empty_params=False
    ):
        """
        Retrieves the URL of the User associated with the speci
        fied Contact Type.
        Args:
            contact_type {str}
        Returns:
            OrgContactUser
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/org/contacts/{contactType}
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OrgContactUser)

        if error:
            return (None, response, error)

        try:
            result = OrgContactUser(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def update_org_contact_user(
            self, contactType, user_id_string,
            keep_empty_params=False
    ):
        """
        Updates the User associated with the specified Contact
        Type.
        Args:
            contact_type {str}
            {user_id_string}
        Returns:
            OrgContactUser
        """
        http_method = "put".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/org/contacts/{contactType}
            """)

        if isinstance(user_id_string, dict):
            body = user_id_string
        else:
            body = user_id_string.as_dict()
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OrgContactUser)

        if error:
            return (None, response, error)

        try:
            result = OrgContactUser(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def update_org_logo(
            self, file,
            keep_empty_params=False
    ):
        """
        Updates the logo for your organization.
        Args:
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/org/logo
            """)

        body = {}
        headers = {}
        form = {
            "file": file,
        }

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, error)

        response, error = await self._request_executor\
            .execute(request)

        if error:
            return (response, error)

        return (response, None)

    async def get_org_preferences(
            self,
            keep_empty_params=False
    ):
        """
        Gets preferences of your organization.
        Args:
        Returns:
            OrgPreferences
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/org/preferences
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OrgPreferences)

        if error:
            return (None, response, error)

        try:
            result = OrgPreferences(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def hide_okta_ui_footer(
            self,
            keep_empty_params=False
    ):
        """
        Hide the Okta UI footer for all end users of your organ
        ization.
        Args:
        Returns:
            OrgPreferences
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/org/preferences/hideEndUserFooter
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OrgPreferences)

        if error:
            return (None, response, error)

        try:
            result = OrgPreferences(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def show_okta_ui_footer(
            self,
            keep_empty_params=False
    ):
        """
        Makes the Okta UI footer visible for all end users of y
        our organization.
        Args:
        Returns:
            OrgPreferences
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/org/preferences/showEndUserFooter
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OrgPreferences)

        if error:
            return (None, response, error)

        try:
            result = OrgPreferences(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def get_okta_communication_settings(
            self,
            keep_empty_params=False
    ):
        """
        Gets Okta Communication Settings of your organization.
        Args:
        Returns:
            OrgOktaCommunicationSetting
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/org/privacy/oktaCommunication
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OrgOktaCommunicationSetting)

        if error:
            return (None, response, error)

        try:
            result = OrgOktaCommunicationSetting(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def opt_in_users_to_okta_communication_emails(
            self,
            keep_empty_params=False
    ):
        """
        Opts in all users of this org to Okta Communication ema
        ils.
        Args:
        Returns:
            OrgOktaCommunicationSetting
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/org/privacy/oktaCommunication/optIn
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OrgOktaCommunicationSetting)

        if error:
            return (None, response, error)

        try:
            result = OrgOktaCommunicationSetting(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def opt_out_users_from_okta_communication_emails(
            self,
            keep_empty_params=False
    ):
        """
        Opts out all users of this org from Okta Communication
        emails.
        Args:
        Returns:
            OrgOktaCommunicationSetting
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/org/privacy/oktaCommunication/optOut
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OrgOktaCommunicationSetting)

        if error:
            return (None, response, error)

        try:
            result = OrgOktaCommunicationSetting(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def get_org_okta_support_settings(
            self,
            keep_empty_params=False
    ):
        """
        Gets Okta Support Settings of your organization.
        Args:
        Returns:
            OrgOktaSupportSettingsObj
        """
        http_method = "get".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/org/privacy/oktaSupport
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OrgOktaSupportSettingsObj)

        if error:
            return (None, response, error)

        try:
            result = OrgOktaSupportSettingsObj(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def extend_okta_support(
            self,
            keep_empty_params=False
    ):
        """
        Extends the length of time that Okta Support can access
        your org by 24 hours. This means that 24 hours are add
        ed to the remaining access time.
        Args:
        Returns:
            OrgOktaSupportSettingsObj
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/org/privacy/oktaSupport/extend
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OrgOktaSupportSettingsObj)

        if error:
            return (None, response, error)

        try:
            result = OrgOktaSupportSettingsObj(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def grant_okta_support(
            self,
            keep_empty_params=False
    ):
        """
        Enables you to temporarily allow Okta Support to access
        your org as an administrator for eight hours.
        Args:
        Returns:
            OrgOktaSupportSettingsObj
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/org/privacy/oktaSupport/grant
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OrgOktaSupportSettingsObj)

        if error:
            return (None, response, error)

        try:
            result = OrgOktaSupportSettingsObj(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)

    async def revoke_okta_support(
            self,
            keep_empty_params=False
    ):
        """
        Revokes Okta Support access to your organization.
        Args:
        Returns:
            OrgOktaSupportSettingsObj
        """
        http_method = "post".upper()
        api_url = format_url(f"""
            {self._base_url}
            /api/v1/org/privacy/oktaSupport/revoke
            """)

        body = {}
        headers = {}
        form = {}

        request, error = await self._request_executor.create_request(
            http_method, api_url, body, headers, form, keep_empty_params=keep_empty_params
        )

        if error:
            return (None, None, error)

        response, error = await self._request_executor\
            .execute(request, OrgOktaSupportSettingsObj)

        if error:
            return (None, response, error)

        try:
            result = OrgOktaSupportSettingsObj(
                self.form_response_body(response.get_body())
            )
        except Exception as error:
            return (None, response, error)
        return (result, response, None)



from Crypto.PublicKey import RSA
from ast import literal_eval
import jose.jwk as jwk
import jose.jwt as jwt
import time
import uuid
import os


class JWT():
    """
    This class creates a JWT from the Okta Client configuration.
    """

    OAUTH_ENDPOINT = "/oauth2/v1/token"
    HASH_ALGORITHM = "RS256"
    PEM_FORMAT = "PKCS1"
    ONE_HOUR = 1 * 60 * 60
    JWT_OPTIONS = {
        'verify_signature': True,
        'verify_aud': True,
        'verify_iat': True,
        'verify_exp': True,
        'verify_nbf': True,
        'verify_iss': True,
        'verify_sub': True,
        'verify_jti': True,
        'verify_at_hash': True,
        'require_aud': True,
        'require_iat': False,
        'require_exp': True,
        'require_nbf': False,
        'require_iss': True,
        'require_sub': True,
        'require_jti': True,
    }

    def __init__(self, private_key):
        pass

    @staticmethod
    def get_PEM_JWK(private_key):
        """
        This class gets the PEM and JWK representation of the private key
        from the Okta Client configuration.

        Args:
            private_key (str or dict): Either a dict representing a JWK, str
            representing the JWK, filepath to PEM, or PEM string

        Returns:
            RSAKey, JWK: Tuple of the generated PEM, and JWK
        """
        # Start off with both as None
        my_jwk = None
        my_pem = None

        # check if JWK
        # String representation of dictionary or dict
        if ((type(private_key) == str and private_key.startswith("{")) or
                type(private_key) == dict):
            # if string repr, convert to dict object
            if (type(private_key) == str):
                private_key = literal_eval(private_key)
            # Create JWK using dict obj
            my_jwk = jwk.construct(private_key, JWT.HASH_ALGORITHM)
        else:  # it's a PEM
            # check for filepath or explicit private key
            if os.path.exists(private_key):
                # open file if exists and import key
                pem_file = open(private_key, 'r')
                my_pem = RSA.import_key(pem_file.read())
                pem_file.close()
            else:
                # convert given string to bytes and import key
                private_key_bytes = bytes(private_key, 'ascii')
                my_pem = RSA.import_key(private_key_bytes)

            if not my_pem:
                # return error if import failed
                return (None, ValueError(
                    "RSA Private Key given is of the wrong type"))

        if my_jwk:  # was JWK provided
            # get PEM using JWK
            pem_bytes = my_jwk.to_pem(JWT.PEM_FORMAT)
            my_pem = RSA.import_key(pem_bytes)
        else:  # was pem provided
            # get JWK using PEM
            my_jwk = jwk.construct(my_pem.export_key(), JWT.HASH_ALGORITHM)

        return (my_pem, my_jwk)

    @staticmethod
    def create_token(org_url, client_id, private_key):
        """
        Create a JSON Web Token using Oauth details from the Okta Client
        config.

        Args:
            org_url (str): Okta Organization URL
            client_id (str): Client ID
            private_key (str or dict): Private Key representation

        Returns:
            str: Generated JWT
        """
        # Generate PEM and JWK
        my_pem, my_jwk = JWT.get_PEM_JWK(private_key)
        # Get current time and expiry time for token
        issued_time = int(time.time())
        expiry_time = issued_time + JWT.ONE_HOUR
        # generate unique JWT ID
        generated_JWT_ID = str(uuid.uuid4())

        # Create claims for token and create token
        claims = {
            'sub': client_id,
            'iat': issued_time,
            'exp': expiry_time,
            'iss': client_id,
            'aud': org_url + JWT.OAUTH_ENDPOINT,
            'jti': generated_JWT_ID
        }

        token = jwt.encode(claims, my_jwk.to_dict(), JWT.HASH_ALGORITHM)
        return token



import pytest
from tests.mocks import MockOktaClient
import okta.models as models
from http import HTTPStatus
from okta.errors.okta_api_error import OktaAPIError


class TestInlineHooksResource:
    """
    Integration Tests for the Inline Hooks Resource
    """
    SDK_PREFIX = "python_sdk"

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_create_get_inline_hook(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Inline Hook
        inline_hook_model = models.InlineHook({
            "name": f"{TestInlineHooksResource.SDK_PREFIX} Test Inline Hook",
            "version": "1.0.0",
            "type": models.InlineHookType.COM_OKTA_OAUTH_2_TOKENS_TRANSFORM,
            "channel": models.InlineHookChannel({
                "type": "HTTP",
                "version": "1.0.0",
                "config": models.InlineHookChannelConfig({
                    "uri": "https://www.example.com/inlineHook",
                    "headers": [
                        models.InlineHookChannelConfigHeaders({
                            "key": "X-Test-Header",
                            "value": "Test Header Value"
                        })
                    ],
                    "authScheme": models.InlineHookChannelConfigAuthScheme({
                        "type": "HEADER",
                        "key": "Authorization",
                        "value": "Test-Api-Key"
                    })
                })
            })
        })

        try:
            created_inline_hook, _, err = await \
                client.create_inline_hook(inline_hook_model)
            assert err is None
            assert isinstance(created_inline_hook, models.InlineHook)
            assert created_inline_hook.id
            assert created_inline_hook.name == inline_hook_model.name
            assert created_inline_hook.channel
            assert created_inline_hook.channel.config
            assert created_inline_hook.channel.config.uri ==\
                inline_hook_model.channel.config.uri

            # Retrieve
            retrieved_inline_hook, _, err = await \
                client.get_inline_hook(created_inline_hook.id)
            assert err is None
            assert isinstance(retrieved_inline_hook, models.InlineHook)
            assert retrieved_inline_hook.id == created_inline_hook.id
            assert retrieved_inline_hook.name == created_inline_hook.name

        finally:
            errors = []
            # Deactivate & Delete
            try:
                deactivated_inline_hook, _, err = await \
                    client.deactivate_inline_hook(created_inline_hook.id)
                assert err is None
                assert isinstance(deactivated_inline_hook, models.InlineHook)
                assert deactivated_inline_hook.status == "INACTIVE"
            except Exception as exc:
                errors.append(exc)

            try:
                _, err = await client.delete_inline_hook(created_inline_hook.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_update_inline_hook(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Inline Hook
        inline_hook_model = models.InlineHook({
            "name": f"{TestInlineHooksResource.SDK_PREFIX} Test Inline Hook",
            "version": "1.0.0",
            "type": models.InlineHookType.COM_OKTA_OAUTH_2_TOKENS_TRANSFORM,
            "channel": models.InlineHookChannel({
                "type": "HTTP",
                "version": "1.0.0",
                "config": models.InlineHookChannelConfig({
                    "uri": "https://www.example.com/inlineHook",
                    "headers": [
                        models.InlineHookChannelConfigHeaders({
                            "key": "X-Test-Header",
                            "value": "Test Header Value"
                        })
                    ],
                    "authScheme": models.InlineHookChannelConfigAuthScheme({
                        "type": "HEADER",
                        "key": "Authorization",
                        "value": "Test-Api-Key"
                    })
                })
            })
        })

        try:
            created_inline_hook, _, err = await \
                client.create_inline_hook(inline_hook_model)
            assert err is None
            assert isinstance(created_inline_hook, models.InlineHook)
            assert created_inline_hook.id
            assert created_inline_hook.name == inline_hook_model.name
            assert created_inline_hook.channel
            assert created_inline_hook.channel.config
            assert created_inline_hook.channel.config.uri ==\
                inline_hook_model.channel.config.uri

            # Update
            updated_inline_hook_model = models.InlineHook({
                "name": inline_hook_model.name + "UPDATE",
                "version": "1.0.0",
                "type": models.InlineHookType.COM_OKTA_OAUTH_2_TOKENS_TRANSFORM,
                "channel": models.InlineHookChannel({
                    "type": "HTTP",
                    "version": "1.0.0",
                    "config": models.InlineHookChannelConfig({
                        "uri": inline_hook_model.channel.config.uri + "UPDATE",
                        "headers": [
                            models.InlineHookChannelConfigHeaders({
                                "key": "X-Test-Header",
                                "value": "Test Header Value UPDATE"
                            })
                        ],
                        "authScheme": models.InlineHookChannelConfigAuthScheme({
                            "type": "HEADER",
                            "key": "Authorization",
                            "value": "Test-Api-Key-UPDATE"
                        })
                    })
                })
            })
            updated_inline_hook, _, err = await \
                client.update_inline_hook(
                    created_inline_hook.id, updated_inline_hook_model)
            assert err is None
            assert updated_inline_hook.id == created_inline_hook.id
            assert updated_inline_hook.name == updated_inline_hook_model.name
            assert updated_inline_hook.channel.config.uri == \
                updated_inline_hook_model.channel.config.uri

            # Retrieve
            retrieved_inline_hook, _, err = await \
                client.get_inline_hook(created_inline_hook.id)
            assert err is None
            assert isinstance(retrieved_inline_hook, models.InlineHook)
            assert retrieved_inline_hook.id == created_inline_hook.id
            assert retrieved_inline_hook.name == updated_inline_hook.name
            assert retrieved_inline_hook.version == created_inline_hook.version
            assert retrieved_inline_hook.type ==\
                models.InlineHookType.COM_OKTA_OAUTH_2_TOKENS_TRANSFORM
            assert retrieved_inline_hook.channel.config.uri ==\
                updated_inline_hook.channel.config.uri

        finally:
            errors = []
            # Deactivate & Delete
            try:
                deactivated_inline_hook, _, err = await \
                    client.deactivate_inline_hook(created_inline_hook.id)
                assert err is None
                assert isinstance(deactivated_inline_hook, models.InlineHook)
                assert deactivated_inline_hook.status == "INACTIVE"
            except Exception as exc:
                errors.append(exc)

            try:
                _, err = await client.delete_inline_hook(created_inline_hook.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_delete_inline_hook(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Inline Hook
        inline_hook_model = models.InlineHook({
            "name": f"{TestInlineHooksResource.SDK_PREFIX} Test Inline Hook",
            "version": "1.0.0",
            "type": models.InlineHookType.COM_OKTA_OAUTH_2_TOKENS_TRANSFORM,
            "channel": models.InlineHookChannel({
                "type": "HTTP",
                "version": "1.0.0",
                "config": models.InlineHookChannelConfig({
                    "uri": "https://www.example.com/inlineHook",
                    "headers": [
                        models.InlineHookChannelConfigHeaders({
                            "key": "X-Test-Header",
                            "value": "Test Header Value"
                        })
                    ],
                    "authScheme": models.InlineHookChannelConfigAuthScheme({
                        "type": "HEADER",
                        "key": "Authorization",
                        "value": "Test-Api-Key"
                    })
                })
            })
        })

        try:
            created_inline_hook, _, err = await \
                client.create_inline_hook(inline_hook_model)
            assert err is None
            assert isinstance(created_inline_hook, models.InlineHook)
            assert created_inline_hook.id
            assert created_inline_hook.name == inline_hook_model.name
            assert created_inline_hook.channel
            assert created_inline_hook.channel.config
            assert created_inline_hook.channel.config.uri ==\
                inline_hook_model.channel.config.uri

            # Deactivate & Delete
            deactivated_inline_hook, _, err = await \
                client.deactivate_inline_hook(created_inline_hook.id)
            assert err is None
            assert isinstance(deactivated_inline_hook, models.InlineHook)
            assert deactivated_inline_hook.status == "INACTIVE"

            _, err = await client.delete_inline_hook(created_inline_hook.id)
            assert err is None

            # Retrieve
            retrieved_inline_hook, resp, err = await \
                client.get_inline_hook(created_inline_hook.id)
            assert err is not None
            assert isinstance(err, OktaAPIError)
            assert resp.get_status() == HTTPStatus.NOT_FOUND
            assert retrieved_inline_hook is None
        finally:
            # Deactivate & Delete
            try:
                _, _, err = await client.deactivate_inline_hook(created_inline_hook.id)
            except Exception:
                pass
            try:
                _, err = await client.delete_inline_hook(created_inline_hook.id)
            except Exception:
                pass

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_activate_deactivate_inline_hook(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        # Create Inline Hook
        inline_hook_model = models.InlineHook({
            "name": f"{TestInlineHooksResource.SDK_PREFIX} Test Inline Hook",
            "version": "1.0.0",
            "type": models.InlineHookType.COM_OKTA_OAUTH_2_TOKENS_TRANSFORM,
            "channel": models.InlineHookChannel({
                "type": "HTTP",
                "version": "1.0.0",
                "config": models.InlineHookChannelConfig({
                    "uri": "https://www.example.com/inlineHook",
                    "headers": [
                        models.InlineHookChannelConfigHeaders({
                            "key": "X-Test-Header",
                            "value": "Test Header Value"
                        })
                    ],
                    "authScheme": models.InlineHookChannelConfigAuthScheme({
                        "type": "HEADER",
                        "key": "Authorization",
                        "value": "Test-Api-Key"
                    })
                })
            })
        })

        try:
            created_inline_hook, _, err = await \
                client.create_inline_hook(inline_hook_model)
            assert err is None
            assert isinstance(created_inline_hook, models.InlineHook)
            assert created_inline_hook.id
            assert created_inline_hook.name == inline_hook_model.name
            assert created_inline_hook.channel
            assert created_inline_hook.channel.config
            assert created_inline_hook.channel.config.uri ==\
                inline_hook_model.channel.config.uri
            assert created_inline_hook.status == "ACTIVE"

            # Deactivate
            deactivated_inline_hook, _, err = await \
                client.deactivate_inline_hook(created_inline_hook.id)
            assert err is None
            assert isinstance(deactivated_inline_hook, models.InlineHook)
            assert deactivated_inline_hook.status == "INACTIVE"

            # Retrieve
            retrieved_inline_hook, _, err = await \
                client.get_inline_hook(created_inline_hook.id)
            assert err is None
            assert retrieved_inline_hook.status == "INACTIVE"

            # Activate
            activated_inline_hook, _, err = await \
                client.activate_inline_hook(created_inline_hook.id)
            assert err is None
            assert isinstance(activated_inline_hook, models.InlineHook)
            assert activated_inline_hook.status == "ACTIVE"

            # Retrieve
            retrieved_inline_hook, _, err = await \
                client.get_inline_hook(created_inline_hook.id)
            assert err is None
            assert retrieved_inline_hook.status == "ACTIVE"

        finally:
            errors = []
            # Delete & Deactivate
            try:
                _, _, err = await \
                    client.deactivate_inline_hook(created_inline_hook.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            try:
                _, err = await client.delete_inline_hook(created_inline_hook.id)
                assert err is None
            except Exception as exc:
                errors.append(exc)
            assert len(errors) == 0

    @pytest.mark.vcr()
    @pytest.mark.asyncio
    async def test_list_inline_hooks(self, fs):
        # Instantiate Mock Client
        client = MockOktaClient(fs)

        try:
            # Create Inline Hooks
            NUMBER_OF_HOOKS = 3
            created_inline_hooks = []
            for index in range(NUMBER_OF_HOOKS):
                inline_hook_model = models.InlineHook({
                    "name": f"{TestInlineHooksResource.SDK_PREFIX} IH{index}",
                    "version": "1.0.0",
                    "type": models.InlineHookType.
                    COM_OKTA_OAUTH_2_TOKENS_TRANSFORM,
                    "channel": models.InlineHookChannel({
                        "type": "HTTP",
                        "version": "1.0.0",
                        "config": models.InlineHookChannelConfig({
                            "uri": "https://www.example.com/inlineHook",
                            "headers": [
                                models.InlineHookChannelConfigHeaders({
                                    "key": "X-Test-Header",
                                    "value": f"Test Header {index}"
                                })
                            ],
                            "authScheme": models.
                            InlineHookChannelConfigAuthScheme({
                                "type": "HEADER",
                                "key": "Authorization",
                                "value": f"Test-Api-Key-{index}"
                            })
                        })
                    })
                })

                created_inline_hook, _, err = await \
                    client.create_inline_hook(inline_hook_model)
                assert err is None
                created_inline_hooks.append(created_inline_hook.id)
                assert isinstance(created_inline_hook, models.InlineHook)
                assert created_inline_hook.id
                assert created_inline_hook.name == inline_hook_model.name
                assert created_inline_hook.channel
                assert created_inline_hook.channel.config
                assert created_inline_hook.channel.config.uri ==\
                    inline_hook_model.channel.config.uri

            # List
            all_inline_hooks, _, err = await client.list_inline_hooks()
            assert err is None
            assert isinstance(all_inline_hooks, list)
            for inline_hook_id in created_inline_hooks:
                assert next((ih for ih in all_inline_hooks
                             if ih.id == inline_hook_id))

        finally:
            errors = []
            # Deactivate & Delete
            for inline_hook_id in created_inline_hooks:
                try:
                    _, _, err = await \
                        client.deactivate_inline_hook(inline_hook_id)
                    assert err is None
                except Exception as exc:
                    errors.append(exc)
                try:
                    _, err = await client.delete_inline_hook(inline_hook_id)
                    assert err is None
                except Exception as exc:
                    errors.append(exc)
            assert len(errors) == 0
